import boto3


def get_arn(cluster_name: str):
    """
    get ARN of a cluster

    :param str cluster_name: cluster name
    :return: ARN of the cluster
    :rtype: str
    """

    client = boto3.client('ecs')
    response = client.describe_clusters(clusters=[cluster_name])
    assert len(response["clusters"]) != 0, "Cluster '{}' not found".format(cluster_name)
    assert len(response["clusters"]) == 1, "There are clusters with the same name '{}'".format(cluster_name)
    return response["clusters"][0]["clusterArn"]


def get_scheduled_task_names(cluster_name: str):
    """
    get scheduled task names in cluster

    :param str cluster_name: cluster name
    :return: scheduled_task_names
    :rtype: list
    """

    client = boto3.client('events')
    return client.list_rule_names_by_target(TargetArn=get_arn(cluster_name))["RuleNames"]



import boto3, tarfile, mimetypes

from typing import Callable
from io import BytesIO


def read_object_string(bucket: str, s3_path_file: str) -> str:
    with BytesIO(get_object(bucket, s3_path_file)['Body'].read()) as fin:
        return fin.read().decode("utf-8")


def get_object(bucket: str, s3_path_file: str):
    return boto3.resource('s3').meta.client.get_object(Bucket=bucket, Key=s3_path_file)


def upload_file(source_file: str, bucket: str, path_to_desttionation_file: str):
    mime_type, _ = mimetypes.guess_type(source_file)
    if mime_type:
        boto3.resource('s3').meta.client.upload_file(source_file, bucket, path_to_desttionation_file, ExtraArgs={'ContentType': mime_type})
    else:
        boto3.resource('s3').meta.client.upload_file(source_file, bucket, path_to_desttionation_file)


def put_dir(bucket: str, s3_path_dir: str, **kw):
    if not s3_path_dir.endswith("/"):
        raise Exception(f"`s3_path_dir` argument do not end with `/`. Got `{s3_path_dir}`")
   
    boto3.resource('s3').meta.client.put_object(Bucket=bucket, Key=s3_path_dir, **kw)


def put_string_file(bucket: str, s3_path_file: str, file_str: str, **kw):
    boto3.resource('s3').meta.client.put_object(Bucket=bucket, Key=s3_path_file, Body=file_str.encode("utf-8"), **kw)
    

def upload_tar_file_extracted_on_the_fly(tar_file: tarfile.TarFile, bucket: str, file_transfer_callback: Callable = None) -> None:
    """
    Upload a tar file and extracted on the fly

    :param tarfile.TarFile tar_file: tar file or gzip tar file
    :param str bucket: S3 bucket
    :param Callable file_transfer_callback: A method which takes a number of bytes transferred to be periodically called during the copy.
    """
    for tar_info in tar_file:
        if tar_info.isdir():
            put_dir(bucket, f"{tar_info.name[2:]}/") # "./build/assets/"

        else:
            mime_type, _ = mimetypes.guess_type(tar_info.name)
            if mime_type:
                boto3.resource('s3').meta.client.upload_fileobj(
                    Fileobj=BytesIO(tar_file.extractfile(tar_info).read()),
                    Bucket=bucket,
                    Key=tar_info.name[2:], # "./build/assets/temp7.png"
                    Callback=file_transfer_callback,
                    ExtraArgs={'ContentType': mime_type}
                )
            else:
                boto3.resource('s3').meta.client.upload_fileobj(
                    Fileobj=BytesIO(tar_file.extractfile(tar_info).read()),
                    Bucket=bucket,
                    Key=tar_info.name[2:], # "./build/assets/temp7.png"
                    Callback=file_transfer_callback
                )


def download_file(desttionation_file: str, bucket: str, path_to_source_file: str):
    boto3.resource('s3').meta.client.download_file(bucket, path_to_source_file, desttionation_file)


def empty_bucket(bucket: str):
    # S3:DeleteObject
    # S3:DeleteObjectVersion
    s3 = boto3.resource('s3')
    bucket_versioning = s3.BucketVersioning(bucket)
    if bucket_versioning.status == 'Enabled':
        return s3.Bucket(bucket).object_versions.delete()
    else:
        return s3.Bucket(bucket).objects.all().delete()


def list(bucket: str, s3_path_file: str) -> list[str]:
    response = boto3.resource('s3').meta.client.list_objects_v2(Bucket=bucket, Prefix=s3_path_file)
    if response["KeyCount"] == 0:
        return []
    return [x["Key"] for x in response["Contents"]]


def is_file_existed(bucket: str, s3_path_file: str) -> bool:
    results = list(bucket, s3_path_file)
    if len(results) > 1:
        raise Exception(f"There are results when list S3 bucket `{bucket}` with prefix `{s3_path_file}`")
    elif len(results) == 1:
        if results[0] != s3_path_file:
            raise Exception(f"Find a result `{results[0]}` with different key when list S3 bucket `{bucket}` with prefix `{s3_path_file}`")
        return True
    else:
        return False


def wait_until_file_existed(bucket: str, s3_path_file: str, wait_times: int, wait_interval):
    waiter = boto3.resource('s3').meta.client.get_waiter('object_exists')
    waiter.wait(
        Bucket=bucket,
        Key=s3_path_file,
        WaiterConfig={
            "Delay": wait_interval,
            "MaxAttempts": wait_times
        }
    )



import boto3


def is_resource(the_arn: str):
    """
    is the ARN is a secret manager resource?

    :param str the_arn: the ARN
    :return: is the ARN is a secret manager resource?
    :rtype: bool
    """

    # arn:aws:secretsmanager:ap-northeast-1:038528481894:secret:ep/ECS-secret/test-XlIeIr
    return the_arn.startswith("arn:aws:secretsmanager:")


def is_reference_existed(reference_arn: str):
    """
    is reference existed?

    :param str reference_arn: reference ARN
    :return: is reference existed?
    :rtype: bool
    """

    client = boto3.client('secretsmanager')
    secrets = client.list_secrets()["SecretList"]
    return reference_arn in [secret["ARN"] for secret in secrets]



from typing import List

import boto3, enum, time


@enum.unique
class DEPLOY_TYPE(enum.Enum):
    CODEDEPLOY = 1
    ROLLING_UPDATE = 2


def get_detail(cluster_name: str, service_name: str) -> dict:
    """
    get service detail

    :param str cluster_name: ECS cluster
    :param str service_name: ECS service
    :return: service
    :rtype: dict
    """

    client = boto3.client('ecs')
    response = client.describe_services(cluster=cluster_name, services=[service_name])
    assert len(response["services"]) == 1, "There should be only one service '{}' in Cluster '{}'. Got '{}'".format(service_name, cluster_name, response["services"])
    return response["services"][0]


def get_all_lb_dns_names(cluster_name: str, service_name: str) -> List[str]:
    """
    get service all load balancer DNS names

    :param str cluster_name: ECS cluster
    :param str service_name: ECS service
    :return: load balancer DNS names
    :rtype: List[str]
    """

    service = get_detail(cluster_name, service_name)
    target_group_arns = [
        lb["targetGroupArn"]
        for task_set in service["taskSets"]
        for lb in task_set["loadBalancers"]
    ]

    client = boto3.client('elbv2')
    response = client.describe_target_groups(TargetGroupArns=target_group_arns)

    if len(response["TargetGroups"]) != len(target_group_arns):
        return Exception(f"Unable to get exaclt {len(target_group_arns)} target group for ARNs `{target_group_arns}`")

    response = client.describe_load_balancers(
        LoadBalancerArns=[
            lb_arn
            for target_group in response["TargetGroups"]
            for lb_arn in target_group["LoadBalancerArns"]
        ]
    )
    return [x["DNSName"] for x in response["LoadBalancers"]]


def get_task_definition_with_rev(cluster_name: str, service_name: str) -> str:
    """
    get service task definition with revision

    :param str cluster_name: ECS cluster
    :param str service_name: ECS service
    :return: task_definition_with_rev
    :rtype: str
    """

    service = get_detail(cluster_name, service_name)
    return service["taskDefinition"].split('/')[1]


def get_deploy_type(cluster_name: str, service_name: str) -> DEPLOY_TYPE:
    """
    get service task definition with revision

    :param str cluster_name: ECS cluster
    :param str service_name: ECS service
    :return: deploy_type
    :rtype: DEPLOY_TYPE
    """

    service = get_detail(cluster_name, service_name)
    if "deploymentController" in service:
        assert service["deploymentController"]["type"] == "CODE_DEPLOY", "Unsupport deployment type for '{}' cluster '{}' service for type '{}'".format(cluster_name, service_name, service["deploymentController"]["type"])
        return DEPLOY_TYPE.CODEDEPLOY
    else:
        return DEPLOY_TYPE.ROLLING_UPDATE


def update_service(cluster_name: str, service_name: str, task_definition_with_rev=None, auto_rollback=False, wait_interval=5, wait_times=120, re_deploy=False) -> None:
    """
    get service task definition with revision

    :param str cluster_name: ECS cluster
    :param str service_name: ECS service
    :param str task_definition_with_rev: task definition with revision
    :param bool auto_rollback: rollback when failed
    :param int wait_interval: wait retry interval
    :param int wait_times: wait times
    """

    if re_deploy:
        assert task_definition_with_rev is None, "Re-deploy should not give task_definition_rev"

        kw = {
            "forceNewDeployment": True
        }
    else:
        kw = {
            "taskDefinition": task_definition_with_rev
        }

    client = boto3.client('ecs')
    response = client.update_service(
        cluster=cluster_name,
        service=service_name,
        deploymentConfiguration={
            "deploymentCircuitBreaker": {
                # The deployment circuit breaker determines whether a service deployment will fail if the service can't reach a steady state. 
                # If deployment circuit breaker is enabled, a service deployment will transition to a failed state and stop launching new tasks. 
                # If rollback is enabled, when a service deployment fails, the service is rolled back to the last deployment that completed successfully.
                "enable": True,
                "rollback": auto_rollback
            }
        },
        healthCheckGracePeriodSeconds=10,
        **kw
    )
    # TODO Check response

    waiter = client.get_waiter('services_stable')
    waiter.wait(cluster=cluster_name, services=[service_name], WaiterConfig={'Delay': wait_interval, 'MaxAttempts': wait_times})
    time.sleep(20)

    service = get_detail(cluster_name, service_name)
    cur_task_definition_with_rev = service["taskDefinition"].split('/')[1]
    assert cur_task_definition_with_rev == task_definition_with_rev, f"Task definition not update to `{task_definition_with_rev}` for cluster `{cluster_name}` service `{service_name}`. Got `{cur_task_definition_with_rev}`"

    results = [x for x in service["deployments"] if x["status"] == "PRIMARY"]
    assert len(results) == 1, f"There is no Primary deployment for cluster `{cluster_name}` service `{service_name}`"
    primary_deployment = results[0]
    assert primary_deployment["rolloutState"] == "COMPLETED", f"After waiting, primary deployment still not COMPLETED for cluster `{cluster_name}` service `{service_name}`. Got primarry deployment ```{primary_deployment}```"
    primary_deployment_cur_task_definition_with_rev = primary_deployment["taskDefinition"].split('/')[1]
    assert primary_deployment_cur_task_definition_with_rev == task_definition_with_rev, "After waiting, primary deployment still not update to task definition `{}` for cluster `{}` service `{}`. Got `{}`".format(task_definition_with_rev, cluster_name, service_name, primary_deployment["taskDefinition"])



import json, boto3


def get_arn(task_definition_with_rev: str):
    """
    get task definition ARN

    :param str task_definition_with_rev: task definition with revision
    :return: task definition ARN
    :rtype: str
    """

    client = boto3.client('ecs') 
    return client.describe_task_definition(taskDefinition=task_definition_with_rev)["taskDefinition"]["taskDefinitionArn"]


def get_request_template_dict(task_definition_with_rev: str) -> dict:
    """
    Create request template dict for 'client.register_task_definition(**task_definition_request_dict)'

    :param str task_definition_with_rev: task definition family with revision (default is latest). EX: ted-fargate:22
    :return: get task definition request  template dict
    :rtype: dict
    """

    client = boto3.client('ecs')
    response = client.describe_task_definition(taskDefinition=task_definition_with_rev)

    # {
    #     'taskDefinition': {
    #         'taskDefinitionArn': 'string',
    #         'containerDefinitions': [
    #             ...
    #             'name': 'string',
    #             'image': 'string',
    #             'command': [
    #                 'string',
    #             ],
    #             'environment': [
    #                 {
    #                     'name': 'string',
    #                     'value': 'string'
    #                 },
    #             ],
    #             'secrets': [
    #                 {
    #                     'name': 'string',
    #                     'valueFrom': 'string'
    #                 },
    #             ],

    # Remove not use arguments
    response["taskDefinition"].pop("registeredAt", None)
    response["taskDefinition"].pop("deregisteredAt", None)
    response["taskDefinition"].pop("taskDefinitionArn", None)
    response["taskDefinition"].pop("revision", None)
    response["taskDefinition"].pop("status", None)
    response["taskDefinition"].pop("requiresAttributes", None)
    response["taskDefinition"].pop("compatibilities", None)
    response["taskDefinition"].pop("registeredAt", None)
    response["taskDefinition"].pop("deregisteredAt", None)
    response["taskDefinition"].pop("registeredBy", None)

    return response["taskDefinition"]


def register_task_definition_from_json_file(json_file: str):
    """
    register new revision for a task definition by json file from function 'make_task_definition_from_json_file'

    :param str json_file: json file from function 'make_task_definition_from_json_file'
    :return: the new revision
    :rtype: str, int
    """

    with open(json_file, "r") as fin:
        client = boto3.client('ecs')
        response = client.register_task_definition(**json.load(fin))
        print("New task definition '{}:{}'".format(response["taskDefinition"]["family"], response["taskDefinition"]["revision"]))
        return response["taskDefinition"]["family"], response["taskDefinition"]["revision"]



from typing import List
import boto3


def is_distribution_id_existed(distribution_id: str) -> bool:
    """
    Is distribution id existed

    :param str distribution_id: distribution id
    :return: is distribution id existed?
    :rtype: bool
    """

    client = boto3.client('cloudfront')

    marker = ""
    while response := client.list_distributions(Marker=marker, MaxItems="1"):
        if distribution_id in [x["Id"] for x in response["DistributionList"]["Items"]]:
            return True

        else:
            if response["DistributionList"]["IsTruncated"]:
                marker = response["DistributionList"]["NextMarker"]

            else:
                return False


def invalidate_distribution(distribution_id: str, paths: List[str], request_id: str, wait_interval=5, wait_times=120) -> None:
    """
    Invalidate distribution

    :param str distribution_id: distribution id
    :param list paths: paths
    """

    client = boto3.client('cloudfront')
    response = client.create_invalidation(
        DistributionId=distribution_id,
        InvalidationBatch={
            "Paths": {
                "Quantity": len(paths),
                "Items":    paths
            },
            "CallerReference": request_id 
        }
    )

    client.get_waiter('invalidation_completed').wait(
        DistributionId=distribution_id,
        Id=response["Invalidation"]["Id"],
        WaiterConfig={
            'Delay':       wait_interval,
            'MaxAttempts': wait_times
        }
    )



import boto3


def is_resource(the_arn: str):
    """
    is the ARN is a SSM resource?

    :param str the_arn: the ARN
    :return: is the ARN is a SSM resource?
    :rtype: bool
    """

    # arn:aws:ssm:ap-northeast-1:038528481894:parameter/ep-test-ssh
    return the_arn.startswith("arn:aws:ssm:")


def is_reference_existed(reference_arn: str):
    """
    is reference existed?

    :param str reference_arn: reference ARN
    :return: is reference existed?
    :rtype: bool
    """

    client = boto3.client('ssm')
    # "arn:aws:ssm:ap-northeast-1:038528481894:parameter/ep-test-ssh"
    parameters = client.get_parameters(Names=[reference_arn.split(':parameter/')[1]])["Parameters"]
    return reference_arn in [parameter["ARN"] for parameter in parameters]



from typing import List
import boto3, time, enum

from . import service as ep_service
from . import task_definition as ep_task_definition
from .. import route53 as ep_route53


@enum.unique
class DEPLOYMENT_STATE(enum.Enum):
    STEADY = 1
    INSTALLING = 2
    WAITING_SWITCH_ROUTE = 3
    WAITING_TERMINATING_ORIGINAL_SET = 4


APPLICATION_NAME = lambda cluster, service: f"AppECS-{cluster}-{service}"
DEPLOYGROUP_NAME = lambda cluster, service: f"DgpECS-{cluster}-{service}"


def make_service_appspec_dict(ecs_cluster: str, ecs_service: str, task_definition_with_rev: str, appspec_ver: str):
    """
    make service appspec for codedeploy

    :param str ecs_cluster: cluster name
    :param str ecs_service: service name
    :param str task_definition_with_rev: task definition with revision ex: td:12
    :param str appspec_ver: appspec version
    :return: appspec dict
    :rtype: dict
    """

    service = ep_service.get_detail(ecs_cluster, ecs_service)

    return {
        "version": appspec_ver,
        "Resources": [
            {
                "TargetService": {
                    "Type": "AWS::ECS::Service",
                    "Properties": {
                        "TaskDefinition": ep_task_definition.get_arn(task_definition_with_rev),
                        "LoadBalancerInfo": {
                            # In a ECS service, there is only one ALB allowed.
                            "ContainerName": service["loadBalancers"][0]["containerName"],
                            "ContainerPort": service["loadBalancers"][0]["containerPort"]
                        }
                    }
                }
            }
        ]
    }


def terminate_last_deployment_original_set(ecs_cluster: str, ecs_service: str, wait_interval=5, wait_times=120) -> None:
    """
    Terminate last deployment original set

    :param str ecs_cluster: ECS cluster
    :param str ecs_service: ECS service
    :param int wait_interval: wait retry interval
    :param int wait_times: wait times
    """
    if last_deployment_id := get_last_deployment_id(ecs_cluster, ecs_service):
        status = get_deployment_status(last_deployment_id, ecs_cluster, ecs_service)
        assert status == DEPLOYMENT_STATE.WAITING_TERMINATING_ORIGINAL_SET

        codedeploy_client = boto3.client('codedeploy')

        response = codedeploy_client.continue_deployment(deploymentId=last_deployment_id, deploymentWaitType='TERMINATION_WAIT')
        # TODO Check response

        waiter = codedeploy_client.get_waiter('deployment_successful')
        waiter.wait(deploymentId=last_deployment_id, WaiterConfig={'Delay': wait_interval, 'MaxAttempts': wait_times})
        time.sleep(10)

        status = get_deployment_status(last_deployment_id, ecs_cluster, ecs_service)
        assert status == DEPLOYMENT_STATE.STEADY, f"After terminating last deployment original set, the deployment is not in steady state. Got `{status}`"
    else:
        raise Exception(f"There is no last deployment for ECS cluster `{ecs_cluster}` and ECS service`{ecs_service}`")


def create_deployment(ecs_cluster: str, ecs_service: str, appspec_json_bucket: str, appspec_json_s3_path: str) -> str:
    """
    Create deployment

    :param str ecs_cluster: ECS cluster
    :param str ecs_service: ECS service
    :param str appspec_json_bucket: S3 bucket that stores the appspec.json
    :param str appspec_json_s3_path: S3 path that stores the appspec.json
    :return: deployment ID
    :rtype: str
    """

    codedeploy_client = boto3.client('codedeploy')

    response = codedeploy_client.create_deployment(
        applicationName=APPLICATION_NAME(ecs_cluster, ecs_service),
        deploymentGroupName=DEPLOYGROUP_NAME(ecs_cluster, ecs_service),
        autoRollbackConfiguration={'enabled': False},
        revision={
            'revisionType': 'S3',
            's3Location': {
                'bucket': appspec_json_bucket,
                'key': appspec_json_s3_path,
                'bundleType': 'JSON'
            }
        }
    )
    return response["deploymentId"]


def switch_route(deployment_id: str, ecs_cluster: str, ecs_service: str, wait_interval=5, wait_times=120) -> None:
    """
    Switch route

    :param str deployment_id: deployment ID
    :param str ecs_cluster: ECS cluster
    :param str ecs_service: ECS service
    :param int wait_interval: wait retry interval
    :param int wait_times: wait times
    """

    status = get_deployment_status(deployment_id, ecs_cluster, ecs_service)
    assert status == DEPLOYMENT_STATE.WAITING_SWITCH_ROUTE, f":x: Deployment `{deployment_id}` is not ready for switching route. Got `{status}`"
    
    codedeploy_client = boto3.client('codedeploy')

    response = codedeploy_client.continue_deployment(deploymentId=deployment_id, deploymentWaitType='READY_WAIT')
    assert response['ResponseMetadata']['HTTPStatusCode'] == 200, ":x: Deployment `{}` switch route failed. Response status `{}`".format(deployment_id, response['ResponseMetadata']['HTTPStatusCode'])


    def wait_until_termination_inprogress():
        time.sleep(wait_interval)
        for _ in range(0, wait_times):
            if is_deployment_in_terminating_orignal_set_state(deployment_id, ecs_cluster, ecs_service):
                return True
            time.sleep(wait_interval)
        return False


    assert wait_until_termination_inprogress()


def rollback_deployment(deployment_id: str, ecs_cluster: str, ecs_service: str, wait_interval=5, wait_times=120) -> None:
    """
    Rollback deployment

    :param str deployment_id: deployment ID
    :param str ecs_cluster: ECS cluster
    :param str ecs_service: ECS service
    :param int wait_interval: wait retry interval
    :param int wait_times: wait times
    """

    status = get_deployment_status(deployment_id, ecs_cluster, ecs_service)
    assert status == DEPLOYMENT_STATE.WAITING_TERMINATING_ORIGINAL_SET, f":x: Deployment `{deployment_id}` is not in WAITING_TERMINATING_ORIGINAL_SET state. Got `{status}`"
    
    codedeploy_client = boto3.client('codedeploy')

    response = codedeploy_client.stop_deployment(deploymentId=deployment_id, autoRollbackEnabled=True)

    time.sleep(wait_interval)

    for _ in range(0, wait_times):
        # It will create a new rollback deployment
        if rollback_deployment_id := get_last_deployment_id(ecs_cluster, ecs_service):
            response = codedeploy_client.get_deployment(deploymentId=rollback_deployment_id)
            if response["deploymentInfo"]["status"] == "Succeeded":
                break
            time.sleep(wait_interval)
        else:
            raise Exception(f"There should be last deployment which is rollback deployment itself for ECS cluster `{ecs_cluster}` and ECS service `{ecs_service}` for rollback source deployment ID `{deployment_id}`")


def get_auto_switch_route_time_str(ecs_cluster: str, ecs_service: str) -> str:
    """
    Get auto switch route time is string format

    :param str ecs_cluster: ECS cluster
    :param str ecs_service: ECS service
    :return: auto switch route time
    :rtype: str
    """

    codedeploy_client = boto3.client('codedeploy')

    response = codedeploy_client.get_deployment_group(
        applicationName=APPLICATION_NAME(ecs_cluster, ecs_service),
        deploymentGroupName=DEPLOYGROUP_NAME(ecs_cluster, ecs_service)
    )
    wait_time = response["deploymentGroupInfo"]["blueGreenDeploymentConfiguration"]["deploymentReadyOption"]
    auto_switch_route_time_total_mins = wait_time.get("waitTimeInMinutes", 0)
    auto_switch_route_time_total_hours = int(auto_switch_route_time_total_mins) / 60
    return "{:.0f} days {:.0f} hours {:.0f} mins".format(auto_switch_route_time_total_hours / 24, auto_switch_route_time_total_hours % 24, int(auto_switch_route_time_total_mins) % 60)


def wait_until_ready(deployment_id: str, wait_times: int, wait_interval: int):
    """
    wait and determine if the new deployment status is ready to switch route

    :param str deployment_id: codedeploy deployment id
    :param int wait_times: waiting max times
    :param int wait_interval: waiting interval
    :return: latest deployment is now ready
    :rtype: boolean
    """

    codedeploy_client = boto3.client('codedeploy')

    for _ in range(0, wait_times):
        time.sleep(wait_interval)
        response = codedeploy_client.get_deployment(deploymentId=deployment_id)
        if response['deploymentInfo']['status'] == 'Ready':
            return
        
    if "errorInformation" in response['deploymentInfo']:
        return response['deploymentInfo']["errorInformation"]["message"]
    else:
        return f"Deployment status in `{response['deploymentInfo']['status']}` without reason"


def wait_until_installation_steady(deployment_id: str, ecs_cluster: str, ecs_service: str, wait_times: int, wait_interval: int):
    """
    wait and determine if the new deployment is in steady state

    :param str deployment_id: codedeploy deployment id
    :param str ecs_cluster: ECS cluster
    :param str ecs_service: ECS service
    :param int wait_times: waiting max times
    :param int wait_interval: waiting interval
    :return: is_finally_in_steady_state
    :rtype: bool
    """

    for _ in range(0, wait_times): # wait at most 10 mins(30sec * 20)
        service = ep_service.get_detail(ecs_cluster, ecs_service)
        results = [x for x in service["taskSets"] if "externalId" in x and x["externalId"] == deployment_id]
        if len(results) != 1:
            raise Exception(f"There should be only 1 deployment task set for `{ecs_cluster}` cluster `{ecs_service}` service ... but got '{results}'")
        task_set = results[0]
        if task_set["stabilityStatus"] == "STEADY_STATE":
            return
        time.sleep(wait_interval)

    events_str =      "\n".join([f"{event['id']}, {event['createdAt']}, {event['message']}" for event in service["events"][:5]])
    deployments_str = "\n".join([f"{deployment['id']}, {deployment['status']}, {deployment['rolloutStateReason']}, {deployment['rolloutState']}" for deployment in service["deployments"]])

    if deployments_str:
        return f"Deployments Status\n{deployments_str}\nRecent 5 events\n{events_str}"
    else:
        return f"Recent 5 events\n{events_str}"


def get_last_deployment_id(ecs_cluster: str, ecs_service: str):
    """
    get last deployment id

    :param str ecs_cluster: ECS cluster
    :param str ecs_service: ECS service
    :return: last deployment id
    :rtype: str
    """

    codedeploy_client = boto3.client('codedeploy')

    response = codedeploy_client.get_deployment_group(
        applicationName=APPLICATION_NAME(ecs_cluster, ecs_service),
        deploymentGroupName=DEPLOYGROUP_NAME(ecs_cluster, ecs_service)
    )
    if "lastAttemptedDeployment" in response['deploymentGroupInfo']:
        return response['deploymentGroupInfo']['lastAttemptedDeployment']['deploymentId']
    else:
        return None


def get_deployment_status(deployment_id: str, ecs_cluster: str, ecs_service: str):
    """
    get deployment status

    :param str deployment_id: deployment id
    :param str ecs_cluster: ECS cluster
    :param str ecs_service: ECS service
    :return: deployment status
    :rtype: DEPLOYMENT_STATE
    """

    codedeploy_client = boto3.client('codedeploy')

    response = codedeploy_client.get_deployment(deploymentId=deployment_id)
    status = response['deploymentInfo']['status']
    
    if status == 'Succeeded' or status == 'Failed' or status == 'Stopped':
        return DEPLOYMENT_STATE.STEADY

    elif status == 'Created' or status == 'Queued' or status == 'Baking':
        return DEPLOYMENT_STATE.INSTALLING

    elif status == 'Ready':
        return DEPLOYMENT_STATE.WAITING_SWITCH_ROUTE

    elif status == 'InProgress':
        if is_deployment_in_terminating_orignal_set_state(deployment_id, ecs_cluster, ecs_service):
            return DEPLOYMENT_STATE.WAITING_TERMINATING_ORIGINAL_SET
        else:
            return DEPLOYMENT_STATE.INSTALLING

    else:
        raise Exception("Exception for un-handling deployment status '{}'".format(status))


def is_deployment_in_terminating_orignal_set_state(deployment_id: str, ecs_cluster: str, ecs_service: str):
    """
    terminating original set state means all deployment lifecycle events succeeded

    :param str deployment_id: deployment id
    :param str ecs_cluser: cluster name
    :param str ecs_service: service name
    :return all deployment lifecycle events succeeded
    :rtype booleen
    """

    codedeploy_client = boto3.client('codedeploy')

    response = codedeploy_client.get_deployment_target(
        deploymentId= deployment_id,
        targetId='{}:{}'.format(ecs_cluster,ecs_service)
    )

    events = response["deploymentTarget"]["ecsTarget"]["lifecycleEvents"]
    not_succeeded_events = [event for event in events if event["status"] != 'Succeeded']
    for not_succeeded_event in not_succeeded_events:
        print('Deployment is still in stage {}'.format(not_succeeded_event["lifecycleEventName"]))
    return bool(not_succeeded_events == [])


def cancel_deployment(deployment_id: str):
    """
    :param str deployment_id: deployment id
    """

    codedeploy_client = boto3.client('codedeploy')

    response = codedeploy_client.stop_deployment(deploymentId=deployment_id, autoRollbackEnabled=False)
    # TODO Check response
    while True:
        response = codedeploy_client.get_deployment(deploymentId=deployment_id)
        status = response['deploymentInfo']['status']
        if  status == 'Stopped':
            return
        time.sleep(2)


def get_blue_set_test_domains(ecs_cluster: str, ecs_service: str) -> List[str]:
    """
    Get blue set load balancer test domains include host and port

    :param str ecs_cluster: ECS cluster
    :param str ecs_service: ECS service
    :return: test domains include host and port
    :rtype: List[str]
    """

    codedeploy_client = boto3.client('codedeploy')

    response = codedeploy_client.get_deployment_group(
        applicationName=APPLICATION_NAME(ecs_cluster, ecs_service),
        deploymentGroupName=DEPLOYGROUP_NAME(ecs_cluster, ecs_service)
    )

    blue_green_pairs = response["deploymentGroupInfo"]["loadBalancerInfo"]["targetGroupPairInfoList"]
    if len(blue_green_pairs) != 1:
        raise Exception(f"There should be only 1 blue green pair. Got\n{blue_green_pairs}")
    blue_green_pair = blue_green_pairs[0]

    if len(blue_green_pair["testTrafficRoute"]["listenerArns"]) != 1:
        raise Exception(f"There should be only 1 blue listener ARN. Got\n{blue_green_pair['testTrafficRoute']['listenerArns']}")
    blue_listener_arn = blue_green_pair["testTrafficRoute"]["listenerArns"][0]

    client = boto3.client('elbv2')
    response = client.describe_listeners(ListenerArns=[blue_listener_arn])

    if len(response["Listeners"]) != 1:
        raise Exception(f"There should be only 1 listener detail for ARN '{blue_listener_arn}'. Got\n{response['Listeners']}")
    test_port = response["Listeners"][0]["Port"]

    lb_dns_names = ep_service.get_all_lb_dns_names(ecs_cluster, ecs_service)
    test_domains = ep_route53.get_record_names_by_alias_target_dns_names(lb_dns_names)
    if test_domains:
        return [f"https://{test_domain}:{test_port}" for test_domain in test_domains]
    else:
        return [f"http://{test_domain}:{test_port}" for test_domain in lb_dns_names]



import boto3, re

from . import task_definition as ep_task_definition


def get_detail(scheduled_task_name: str):
    """
    get scheduled task

    :param str scheduled_task_name: scheduled task name
    :return: scheduled_task
    :rtype: dict
    """

    client = boto3.client('events') 
    return client.describe_rule(Name=scheduled_task_name)


def get_targets(scheduled_task_name: str):
    """
    get scheduled task targets

    :param str scheduled_task_name: scheduled task name
    :return: scheduled_task_targets
    :rtype: list
    """

    client = boto3.client('events')
    return client.list_targets_by_rule(Rule=scheduled_task_name)["Targets"]


def get_cron(scheduled_task_name: str) -> str:
    """
    get scheduled task cron

    :param str scheduled_task_name: scheduled task name
    :return: cron
    :rtype: str
    """

    response = get_detail(scheduled_task_name)

    cron_reg = r"^cron\((.+)\)$"
    if m := re.match(cron_reg, response["ScheduleExpression"]):
        return m.group(1)
    raise Exception("Scheduled task `{}` cron not match `{}` got `{}`".format(scheduled_task_name, cron_reg, response["ScheduleExpression"]))


def get_enable(scheduled_task_name: str) -> bool:
    """
    get scheduled task enable

    :param str scheduled_task_name: scheduled task name
    :return: enable
    :rtype: bool
    """

    response = get_detail(scheduled_task_name)

    return True if response["State"] == "ENABLED" else False


def get_target_task_definition_rev(scheduled_task_name: str, scheduled_task_target_name: str) -> str:
    """
    get scheduled task target task definition with revision

    :param str scheduled_task_name: scheduled task name
    :param str scheduled_task_target_name: scheduled task target name
    :return: scheduled task target task definition with revision
    :rtype: str
    """

    response = get_targets(scheduled_task_name)

    results = [target["EcsParameters"]['TaskDefinitionArn'].split('/')[1] for target in response if target['Id'] == scheduled_task_target_name]
    assert len(results) == 1, f"Multiple results for schedule task `{scheduled_task_name}` and target `{scheduled_task_target_name}`"
    return results[0]


def update(scheduled_task_name: str, cron_exp=None, enable=None):
    """
    update scheduled task

    :param str scheduled_task_name: scheduled task name
    :param str cron_exp: EX: '0 11 * * ? *'
    :param str enable: enable scheduled task?
    :return: response
    :rtype: dict
    """
    # Fixed interval "rate(1 hour)" not support

    args = {}
    if enable is not None:
        args["State"] = "ENABLED" if enable else "DISABLED"

    original_scheduled_task = get_detail(scheduled_task_name)

    client = boto3.client('events') 
    return client.put_rule(
        Name=scheduled_task_name,
        ScheduleExpression="cron({})".format(cron_exp) if cron_exp else original_scheduled_task["ScheduleExpression"],
        **args
    )


def update_target(scheduled_task_name: str, update_target_id: str, update_task_definition_with_rev: str):
    """
    update a scheduled task target
    沒列在 scheduled_task_target_task_definition 內的 target 但是 scheduled task 既有的 target 會保留
    要刪除請實作 remove_targets()

    :param str scheduled_task_name: scheduled task name
    :param str request_update_target_id: update target id
    :param str update_task_definition_with_rev: update target task definition with revision
    :return: response
    :rtype: dict
    """

    scheduled_targets = get_targets(scheduled_task_name)
    assert update_target_id in [x["Id"] for x in scheduled_targets]
    for target in scheduled_targets:
        if target["Id"] == update_target_id:
            assert "EcsParameters" in target, "Update target '{}' but target without 'EcsParameters'".format(target["Id"])
            target["EcsParameters"]["TaskDefinitionArn"] = ep_task_definition.get_arn(update_task_definition_with_rev)
 
    client = boto3.client('events') 
    return client.put_targets(Rule=scheduled_task_name, Targets=scheduled_targets)



from typing import List

import boto3


def get_record_names_by_alias_target_dns_names(alias_target_dns_names: List[str]) -> List[str]:
    """
    Get record names by alias target DNS names

    :param List[str] alias_target_dns_names: alias target DNS names
    :return: record names
    :rtype: List[str]
    """

    search_strs = [f"{x}." for x in alias_target_dns_names]

    client = boto3.client('route53')
    response = client.list_hosted_zones()
    results = []
    for hosted_zone_id in (x["Id"] for x in response["HostedZones"]):
        response = client.list_resource_record_sets(HostedZoneId=hosted_zone_id)
        for record_set in (x for x in response["ResourceRecordSets"] if x["Type"] == "A"):
            if record_set["AliasTarget"]["DNSName"] in search_strs:
                results.append(record_set["Name"][:-1])
    return results



import boto3


def get_image_uri(repo_name: str) -> str:
    """
    get image uri

    :param str repo_name: ecr repo name
    :return: image uri
    :rtype: str
    """

    client = boto3.client('ecr')
    response = client.describe_repositories(
        repositoryNames=[repo_name]
    )

    assert len(response["repositories"]) == 1, "There should be only one ECR repository but got '{}'".format(response)
    return response["repositories"][0]["repositoryUri"]


def get_image_registry(repo_name: str) -> str:
    """
    get image registry

    :param str repo_name: ecr repo name
    :return: image registry
    :rtype: str
    """

    client = boto3.client('ecr')
    response = client.describe_repositories(
        repositoryNames=[repo_name]
    )

    assert len(response["repositories"]) == 1, "There should be only one ECR repository but got '{}'".format(response)
    return response["repositories"][0]["repositoryUri"].split("/")[0]


def is_registry(registry: str):
    """
    is the ecr registry

    :param str registry: ECR registry
    :return: is the registry exist
    :rtype: boolean
    """

    client = boto3.client('ecr')
    response = client.describe_registry()
    return registry == "{}.dkr.ecr.{}.amazonaws.com".format(response["registryId"], client.meta.region_name)


def has_repo(repo_name: str):
    """
    does ecr image exist in ecr

    :param str repo_name: ecr repo name
    :return: repo exist in ecr
    :rtype: boolean
    """

    client = boto3.client('ecr')
    try:
        client.describe_repositories(repositoryNames=[repo_name])
        return True
    except client.exceptions.RepositoryNotFoundException:
        return False


def does_repo_has_tag(repo_name: str, tag: str):
    """
    does ecr repo exist with tag

    :param str repo_name: ecr repo name
    :return: repo with tag exist in ecr 
    :rtype: boolean
    """
    
    client = boto3.client('ecr')
    response = client.list_images(repositoryName=repo_name)
    return tag in [x["imageTag"] for x in response["imageIds"]]



import boto3, functools


def assume_role(assume_role_arn: str, assume_role_session_name: str, assume_role_region: str, external_id: str):
    def _decorator(func):
        @functools.wraps(func)
        def _wrapper(*args, **kw):

            print(f"Assume role `{assume_role_arn}`")
            print(f"Before assuming role, the identity is `{get_caller_identity()}`")

            assumed_role = boto3.client('sts').assume_role(
                RoleArn =         assume_role_arn,
                RoleSessionName = assume_role_session_name,
                ExternalId=external_id
            )
            boto3.setup_default_session(
                aws_access_key_id     = assumed_role['Credentials']['AccessKeyId'],
                aws_secret_access_key = assumed_role['Credentials']['SecretAccessKey'],
                aws_session_token     = assumed_role['Credentials']['SessionToken'],
                region_name           = assume_role_region
            )

            print(f"After assuming role, the identity is `{get_caller_identity()}`")

            result = func(*args, **kw)

            boto3.setup_default_session()

            return result
        return _wrapper
    return _decorator


def get_caller_identity() -> dict:
    """
    Get caller identity

    :param str reference_arn: reference ARN
    :return: is reference existed?
    :rtype: bool
    """

    return boto3.client("sts").get_caller_identity()



import os

from setuptools import setup, find_packages

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), fname), "r") as fin:
        return fin.read()


setup(
    name = "aws-easy-use",
    version = "0.0.4",
    author = "Ted Chen",
    author_email = "imaging8896@gmail.com",
    description = "AWS boto3 wrapper. Make it easier and intuitive to use.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    license = "MIT",
    url = "https://github.com/imaging8896/aws-easy-use",
    project_urls={
        # 'Documentation': 'https://packaging.python.org/tutorials/distributing-packages/',
        # 'Funding': 'https://donate.pypi.org',
        'Source': "https://github.com/imaging8896/aws-easy-use",
        # 'Tracker': 'https://github.com/pypa/sampleproject/issues',
    },
    packages=find_packages(exclude=["tests*"]),
    python_requires=">=3",
    install_requires=[
        "boto3>=1.24.28"
    ],
    # setup_requires=[
    #     "feedparser>=5.1.3",
    # ],
    # see http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 1 - Planning",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3 :: Only",
    ],
)